package by.zerg.layouteditor.gui;

import static by.zerg.layouteditor.helpers.Constants.*;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.text.Document;
import javax.swing.text.Highlighter;
import javax.swing.text.JTextComponent;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoManager;

import org.apache.log4j.Logger;

import by.zerg.layouteditor.theme.Config;
import by.zerg.layouteditor.theme.Loader;
import by.zerg.layouteditor.theme.Property;
import by.zerg.layouteditor.theme.Section;
import by.zerg.layouteditor.theme.Theme;

public class EditorFrame extends JFrame {
	private static final Logger LOG = Logger.getLogger(EditorFrame.class);

	private Loader loader;

	private JFileChooser fc;
	private JButton loadButton;
	private JButton saveButton;
	private JButton openButton;
	private JButton launchButton;
	private JTextArea configArea;
	private JComboBox themesCombo;
	private JTree configTree;
	private PreviewPanel preview;

	public EditorFrame() {
		fc = new JFileChooser(DEFAULT_PATH);
		fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		fc.setFileFilter(new FileFilter() {

			@Override
			public String getDescription() {
				return POKERSTARS_FOLDER;
			}

			@Override
			public boolean accept(File f) {
				return f.isDirectory();
			}
		});

		ActionListener actionListener = new ActionListener() {
			private void loadThemes() {
				int returnVal = fc.showOpenDialog(null);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					File selected = fc.getSelectedFile();
					try {
						updateTree(null);
						preview.setTheme(null);
						themesCombo.removeAllItems();
						configArea.setText(null);
						configArea.getHighlighter().removeAllHighlights();
						loader = new Loader(selected.getAbsolutePath());
						loader.load();
						for (Theme theme : loader.getThemes()) {
							themesCombo.addItem(theme);
							printTheme(theme);
						}
						configArea.setCaretPosition(0);
					} catch (Throwable t) {
						LOG.error(
								"unhandled exception while loading themes...",
								t);
					}
				}
			}

			private void openTheme(Theme theme) {
				try {
					updateTree(null);
					preview.setTheme(null);
					configArea.setText(null);
					configArea.getHighlighter().removeAllHighlights();
					theme.open();
					Config config = theme.getConfig();
					configArea.setText(config.toString());
					configArea.setCaretPosition(0);
					updateTree(config);
					preview.setTheme(theme);
				} catch (Throwable t) {
					LOG.error("unhandled exception while opening theme...", t);
				}
			}

			private void saveTheme(Theme theme) {
				try {
					// Config config = theme.getConfig();
					// config.setData(configArea.getText());
					// config.encrypt();
					theme.getConfig().setData(configArea.getText());
					theme.save();
					openTheme(theme);
				} catch (Throwable t) {
					LOG.error("unhandled exception while saving theme...", t);
				}
			}

			private void launchClient() {
				try {
					setEnabled(false);
					setState(JFrame.ICONIFIED);
					int exitValue = loader.launchClient();
					if (LOG.isDebugEnabled()) {
						LOG.debug("client was launched with value: "
								+ exitValue);
					}
				} catch (Throwable t) {
					LOG.error("unhandled exception while launching client...",
							t);
				}
				setEnabled(true);
				setExtendedState(JFrame.NORMAL);
			}

			@Override
			public void actionPerformed(ActionEvent event) {
				String actionCommand = event.getActionCommand();
				if (LOAD_ACTION.equals(actionCommand)) {
					loadThemes();
				} else if (OPEN_ACTION.equals(actionCommand)) {
					Theme theme = (Theme) themesCombo.getSelectedItem();
					openTheme(theme);
				} else if (SAVE_ACTION.equals(actionCommand)) {
					Theme theme = (Theme) themesCombo.getSelectedItem();
					saveTheme(theme);
				} else if (LAUNCH_ACTION.equals(actionCommand)) {
					launchClient();
				}
			}
		};

		loadButton = createButton(LOAD_BUTTON_TEXT, LOAD_ACTION, actionListener);
		loadButton.setFont(DEFAULT_FONT);
		openButton = createButton(OPEN_BUTTON_TEXT, OPEN_ACTION, actionListener);
		openButton.setFont(DEFAULT_FONT);
		saveButton = createButton(SAVE_BUTTON_TEXT, SAVE_ACTION, actionListener);
		saveButton.setFont(DEFAULT_FONT);
		launchButton = createButton(LAUNCH_BUTTON_TEXT, LAUNCH_ACTION,
				actionListener);
		launchButton.setFont(DEFAULT_FONT);

		configArea = createTextArea();

		themesCombo = new JComboBox() {
			public boolean isVisible() {
				return super.isVisible() && getItemCount() > 0;
			};
		};
		themesCombo.setFont(DEFAULT_FONT);

		configTree = createTree();

		preview = new PreviewPanel();
	}

	private void printTheme(Theme theme) {
		configArea.append("[" + theme.getId() + "]\nname:\t" + theme.getName()
				+ "\ncopyright:\t" + theme.getCopyright() + "\nbase:\t"
				+ theme.getBaseId() + "\nfolder:\t" + theme.getFolder()
				+ "\ndeck:\t" + theme.getDeck() + "\nfeature:\t"
				+ theme.getFeature() + "\nlobby:\t" + theme.getLobby()
				+ "\npreview:\t" + theme.getPreview() + "\ntype:\t"
				+ (theme.isDefault() ? "default" : "custom") + "\nbaseTheme:\t"
				+ theme.getBaseTheme() + "\ngx.ini:\t"
				+ theme.getConfig().getGx() + "\ngraphics:\t"
				+ theme.getGraphicsFolder() + "\n\n");
	}

	private void highlightConfig(int p1, int p2) {
		try {
			Highlighter highlighter = configArea.getHighlighter();
			highlighter.addHighlight(Math.min(p1, p2), Math.max(p1, p2),
					new Highlighter.HighlightPainter() {

						@Override
						public void paint(Graphics g, int offs0, int offs1,
								Shape bounds, JTextComponent c) {
							try {
								c.repaint();
								g.setColor(Color.RED);
								Rectangle r0 = c.modelToView(offs0);
								Rectangle r1 = c.modelToView(offs1);
								g.drawRect(
										(int) r0.getX()
												+ HIGHLIGHT_BORDER_MARGIN_LEFT,
										(int) r0.getY()
												+ HIGHLIGHT_BORDER_MARGIN_TOP,
										(int) r1.getX()
												+ HIGHLIGHT_BORDER_MARGIN_RIGHT,
										(int) r0.getHeight()
												+ HIGHLIGHT_BORDER_MARGIN_BOTTOM);
							} catch (Throwable t) {
								LOG.error(
										"unhandled exception while painting highlighter...",
										t);
							}
						}

					});
			configArea.setCaretPosition(p1);
		} catch (Throwable t) {
			LOG.error("unhandled exception while painting highlighter...", t);
		}
	}

	private void highlightConfig(Object selected) {
		configArea.getHighlighter().removeAllHighlights();
		if (selected instanceof Section) {
			highlightConfig((Section) selected);
		} else if (selected instanceof Property) {
			highlightConfig((Property) selected);
		}
	}

	private void highlightConfig(Section section) {
		String text = section.toString();
		int p1 = configArea.getText().indexOf(text);
		int p2 = p1 + text.length();
		highlightConfig(p1, p2);
	}

	private void highlightConfig(Property property) {
		int start = configArea.getText().indexOf(
				property.getSection().toString());
		String text = property.getName() + EQUAL;
		int p1 = configArea.getText().indexOf(text, start);
		int p2 = configArea.getText().indexOf(NEW_LINE, p1);
		highlightConfig(p1, p2);
	}

	private void updateTree(Config config) throws IOException {
		DefaultTreeModel treeModel = (DefaultTreeModel) configTree.getModel();
		DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) treeModel
				.getRoot();
		rootNode.removeAllChildren();
		treeModel.reload();
		if (config != null) {
			for (Section section : config.getSections().values()) {
				DefaultMutableTreeNode sectionNode = new DefaultMutableTreeNode(
						section/* .getName() */, true);
				treeModel.insertNodeInto(sectionNode, rootNode,
						rootNode.getChildCount());
				for (Property property : section.getProperties().values()) {
					treeModel.insertNodeInto(new DefaultMutableTreeNode(
							property/* .getName() */, false), sectionNode,
							sectionNode.getChildCount());
				}
			}
		}
	}

	private JButton createButton(String label, String actionCommand,
			ActionListener actionListener) {
		JButton newButton = new JButton(label);
		newButton.setActionCommand(actionCommand);
		newButton.addActionListener(actionListener);
		return newButton;
	}

	private JTextArea createTextArea() {
		JTextArea newTextArea = new JTextArea(CONFIG_TEXTAREA_VISIBLE_ROWS, 0);
		newTextArea.setFont(DEFAULT_FONT);
		newTextArea.setMargin(CONFIG_TEXTAREA_MARGINS);

		final UndoManager undo = new UndoManager();
		Document doc = newTextArea.getDocument();

		// Listen for undo and redo events
		doc.addUndoableEditListener(new UndoableEditListener() {
			public void undoableEditHappened(UndoableEditEvent evt) {
				undo.addEdit(evt.getEdit());
			}
		});

		// Create an undo action and add it to the text component
		newTextArea.getActionMap().put(UNDO_ACTION,
				new AbstractAction(UNDO_ACTION) {
					public void actionPerformed(ActionEvent evt) {
						try {
							if (undo.canUndo()) {
								undo.undo();
							}
						} catch (CannotUndoException e) {
						}
					}
				});

		// Bind the undo action to ctl-Z
		newTextArea.getInputMap().put(KeyStroke.getKeyStroke(CTRL_Z_HOTKEY),
				UNDO_ACTION);

		// Create a redo action and add it to the text component
		newTextArea.getActionMap().put(REDO_ACTION,
				new AbstractAction(REDO_ACTION) {
					public void actionPerformed(ActionEvent evt) {
						try {
							if (undo.canRedo()) {
								undo.redo();
							}
						} catch (CannotRedoException e) {
						}
					}
				});

		// Bind the redo action to ctl-Y
		newTextArea.getInputMap().put(KeyStroke.getKeyStroke(CTRL_Y_HOTKEY),
				REDO_ACTION);
		return newTextArea;
	}

	private JTree createTree() {
		JTree tree = new JTree(new DefaultMutableTreeNode(CONFIG_TREE_ROOT),
				true) {
			@Override
			public Dimension getPreferredScrollableViewportSize() {
				return new Dimension(CONFIG_TREE_DEFAULT_HEIGHT,
						super.getPreferredScrollableViewportSize().height);
			}
		};
		tree.getSelectionModel().setSelectionMode(
				TreeSelectionModel.SINGLE_TREE_SELECTION);
		tree.addTreeSelectionListener(new TreeSelectionListener() {
			@Override
			public void valueChanged(TreeSelectionEvent e) {
				Object selected = ((DefaultMutableTreeNode) e.getPath()
						.getLastPathComponent()).getUserObject();
				highlightConfig(selected);
			}
		});
		DefaultTreeCellRenderer renderer = new CustomTreeCellRenderer();
		tree.setCellRenderer(renderer);
		// DefaultTreeCellEditor editor = new CustomTreeCellEditor(tree,
		// renderer);
		// tree.setCellEditor(editor);
		tree.setEditable(false);
		tree.setToggleClickCount(CONFIG_TREE_TOGGLE_CLICKS);
		tree.setRootVisible(true);
		tree.setFont(DEFAULT_FONT);

		return tree;
	}

	private void init() {
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		setTitle(EDITOR_FRAME_TITLE);
		setLayout(new BorderLayout());

		JPanel topPanel = new JPanel();
		topPanel.add(loadButton);
		topPanel.add(themesCombo);
		topPanel.add(openButton);
		topPanel.add(saveButton);
		topPanel.add(launchButton);
		add(topPanel, BorderLayout.NORTH);

		JPanel bottomPanel = new JPanel(new BorderLayout());
		JPanel centralPanel = new JPanel(new BorderLayout());

		centralPanel.add(preview, BorderLayout.WEST);
		centralPanel.add(new JScrollPane(configTree));
		bottomPanel.add(centralPanel, BorderLayout.NORTH);
		bottomPanel.add(new JScrollPane(configArea));
		add(bottomPanel);

		pack();
		setLocationRelativeTo(null);
		setVisible(true);
	}

	public static void main(String[] args) {
		new EditorFrame().init();
	}
}
